"""HDF5 Tensorflow dataloader.

Convenient way of loading data from hdf5 files in a ML pipeline.
"""

from functools import partial
from typing import List

import keras
import tensorflow as tf
from keras.src.trainers.data_adapters import TFDatasetAdapter

from zea.data.dataloader import H5Generator
from zea.data.layers import Resizer
from zea.utils import find_methods_with_return_type, translate

METHODS_THAT_RETURN_DATASET = find_methods_with_return_type(tf.data.Dataset, "DatasetV2")


class TFDatasetToKeras(TFDatasetAdapter):
    """Tensorflow Dataset to Keras Dataset.

    This class wraps a tf.data.Dataset object and allows it to be used with Keras backends.
    """

    def __init__(self, dataset):
        super().__init__(dataset)

    def __iter__(self):
        backend = keras.backend.backend()
        if backend == "tensorflow":
            return iter(self.get_tf_dataset())
        elif backend == "jax":
            return self.get_jax_iterator()
        elif backend == "torch":
            return iter(self.get_torch_dataloader())
        elif backend == "numpy":
            return self.get_numpy_iterator()
        else:
            raise ValueError(
                f"Unsupported backend: {backend}. "
                "Please use one of the following: 'tensorflow', 'jax', 'torch', 'numpy'."
            )

    def __len__(self):
        return self.num_batches

    def __getattr__(self, name):
        # Delegate all calls to self._dataset, and wraps the result in TFDatasetToKeras
        if name in METHODS_THAT_RETURN_DATASET:

            def method(*args, **kwargs):
                result = getattr(self._dataset, name)(*args, **kwargs)
                return TFDatasetToKeras(result)

            return method
        else:
            return getattr(self._dataset, name)


class H5GeneratorTF(H5Generator):
    """Adds a tensorflow dtype property and output_signature to the H5Generator class."""

    @property
    def tensorflow_dtype(self):
        """
        Extracts one image from the dataset to get the dtype. Converts it to a tensorflow dtype.
        """
        out = next(self.iterator())
        if self.return_filename:
            out = out[0]
        dtype = out.dtype
        if "float" in str(dtype):
            dtype = tf.float32
        elif "complex" in str(dtype):
            dtype = tf.complex64
        elif "uint8" in str(dtype):
            dtype = tf.uint8
        else:
            raise ValueError(f"Unsupported dtype: {dtype}")
        return dtype

    @property
    def output_signature(self):
        """
        Get the output signature of the generator as a tensorflow `TensorSpec`.
        This is useful for creating a `tf.data.Dataset` from the generator.
        """
        output_signature = tf.TensorSpec(shape=self.shape, dtype=self.tensorflow_dtype)
        if self.return_filename:
            output_signature = (
                output_signature,
                tf.TensorSpec(shape=(), dtype=tf.string),
            )
        return output_signature


def _assert_image_range(images, image_range):
    # Check if there are outliers in the image range
    minval = tf.reduce_min(images)
    maxval = tf.reduce_max(images)
    _msg = f"Image range {image_range} is not in the range of the data {minval} - {maxval}"
    tf.debugging.assert_greater_equal(
        minval,
        tf.cast(image_range[0], minval.dtype),
        message=_msg,
    )
    tf.debugging.assert_less_equal(
        maxval,
        tf.cast(image_range[1], maxval.dtype),
        message=_msg,
    )
    return images


def make_dataloader(
    file_paths: List[str],
    batch_size: int,
    key: str = "data/image",
    n_frames: int = 1,
    shuffle: bool = True,
    return_filename: bool = False,
    limit_n_samples: int | None = None,
    limit_n_frames: int | None = None,
    seed: int | None = None,
    drop_remainder: bool = False,
    resize_type: str | None = None,
    resize_axes: tuple | None = None,
    resize_kwargs: dict | None = None,
    image_size: tuple | None = None,
    image_range: tuple | None = None,
    normalization_range: tuple | None = None,
    dataset_repetitions: int | None = None,
    cache: bool = False,
    additional_axes_iter: tuple | None = None,
    sort_files: bool = True,
    overlapping_blocks: bool = False,
    augmentation: callable = None,
    assert_image_range: bool = True,
    clip_image_range: bool = False,
    initial_frame_axis: int = 0,
    insert_frame_axis: bool = True,
    frame_index_stride: int = 1,
    frame_axis: int = -1,
    validate: bool = True,
    prefetch: bool = True,
    shard_index: int | None = None,
    num_shards: int = 1,
    wrap_in_keras: bool = True,
    **kwargs,
) -> tf.data.Dataset:
    """Creates a ``tf.data.Dataset`` from .hdf5 files in the specified directory or directories.

    Mimics the native TF function ``tf.keras.utils.image_dataset_from_directory``
    but for .hdf5 files.

    Saves a dataset_info.yaml file in the directory with information about the dataset.
    This file is used to load the dataset later on, which speeds up the initial loading
    of the dataset for very large datasets.

    Does the following in order to load a dataset:

        - Find all .hdf5 files in the director(ies)
        - Load the data from each file using the specified key
        - Apply the following transformations in order (if specified):

            - limit_n_samples
            - cache
            - shuffle
            - shard
            - add channel dim
            - assert_image_range
            - clip_image_range
            - resize
            - repeat
            - batch
            - normalize
            - augmentation
            - prefetch
            - tf -> keras tensor

    Args:
        file_paths (str or list): Path(s) to the folder(s) or h5 file(s) to load.
        batch_size (int): Batch the dataset.
        key (str): The key to access the HDF5 dataset.
        n_frames (int, optional): Number of frames to load from each hdf5 file.
            Defaults to 1. These frames are stacked along the last axis (channel).
        shuffle (bool, optional): Shuffle dataset.
        return_filename (bool, optional): Return file name with image. Defaults to False.
        limit_n_samples (int, optional): Take only a subset of samples.
            Useful for debugging. Defaults to None.
        limit_n_frames (int, optional): Limit the number of frames to load from each file.
            This means n_frames per data file will be used. These will be the first frames in
            the file. Defaults to None.
        seed (int, optional): Random seed of shuffle.
        drop_remainder (bool, optional): Whether the last batch should be dropped.
        resize_type (str, optional): Resize type. Defaults to 'center_crop'.
            Can be 'center_crop', 'random_crop' or 'resize'.
        resize_axes (tuple, optional): Axes to resize along. Should be of length 2
            (height, width) as resizing function only supports 2D resizing / cropping.
            Should only be set when your data is more than (h, w, c). Defaults to None.
            Note that it considers the axes after inserting the frame axis.
        resize_kwargs (dict, optional): Kwargs for the resize function.
        image_size (tuple, optional): Resize images to image_size. Should
            be of length two (height, width). Defaults to None.
        image_range (tuple, optional): Image range. Defaults to (0, 255).
            Will always translate from specified image range to normalization range.
            If image_range is set to None, no normalization will be done. Note that it does not
            clip to the image range, so values outside the image range will be outside the
            normalization range!
        normalization_range (tuple, optional): Normalization range. Defaults to (0, 1).
            See image_range for more info!
        dataset_repetitions (int, optional): Repeat dataset. Note that this happens
            after sharding, so the shard will be repeated. Defaults to None.
        cache (bool, optional): Cache dataset to RAM.
        additional_axes_iter (tuple, optional): Additional axes to iterate over
            in the dataset. Defaults to None, in that case we only iterate over
            the first axis (we assume those contain the frames).
        sort_files (bool, optional): Sort files by number. Defaults to True.
        overlapping_blocks (bool, optional): If True, blocks overlap by n_frames - 1.
            Defaults to False. Has no effect if n_frames = 1.
        augmentation (keras.Sequential, optional): Keras augmentation layer.
        assert_image_range (bool, optional): Assert that the image range is
            within the specified image range. Defaults to True.
        clip_image_range (bool, optional): Clip the image range to the specified
            image range. Defaults to False.
        initial_frame_axis (int, optional): Axis where in the files the frames are stored.
            Defaults to 0.
        insert_frame_axis (bool, optional): If True, new dimension to stack
            frames along will be created. Defaults to True. In that case
            frames will be stacked along existing dimension (frame_axis).
        frame_index_stride (int, optional): Interval between frames to load.
            Defaults to 1. If n_frames > 1, a lower frame rate can be simulated.
        frame_axis (int, optional): Dimension to stack frames along.
            Defaults to -1. If insert_frame_axis is True, this will be the
            new dimension to stack frames along.
        validate (bool, optional): Validate if the dataset adheres to the zea format.
            Defaults to True.
        prefetch (bool, optional): Prefetch the dataset. Defaults to True.
        shard_index (int, optional): Index which part of the dataset should be selected.
            Can only be used if num_shards is specified. Defaults to None.
            See for info: https://www.tensorflow.org/api_docs/python/tf/data/Dataset#shard
        num_shards (int, optional): This is used to divide the dataset into ``num_shards`` parts.
            Sharding happens before all other operations. Defaults to 1.
            See for info: https://www.tensorflow.org/api_docs/python/tf/data/Dataset#shard
        wrap_in_keras (bool, optional): Wrap dataset in TFDatasetToKeras. Defaults to True.
            If True, will convert the dataset that returns backend tensors.

    Returns:
        tf.data.Dataset: The constructed dataset.

    """
    # Setup
    if normalization_range is not None:
        assert image_range is not None, (
            "If normalization_range is set, image_range must be set as well."
        )

    resize_kwargs = resize_kwargs or {}

    if num_shards > 1:
        assert shard_index is not None, "shard_index must be specified"
        assert shard_index < num_shards, "shard_index must be less than num_shards"
        assert shard_index >= 0, "shard_index must be greater than or equal to 0"

    image_extractor = H5GeneratorTF(
        file_paths,
        key,
        n_frames=n_frames,
        frame_index_stride=frame_index_stride,
        frame_axis=frame_axis,
        insert_frame_axis=insert_frame_axis,
        initial_frame_axis=initial_frame_axis,
        return_filename=return_filename,
        shuffle=shuffle,
        sort_files=sort_files,
        overlapping_blocks=overlapping_blocks,
        limit_n_samples=limit_n_samples,
        limit_n_frames=limit_n_frames,
        seed=seed,
        additional_axes_iter=additional_axes_iter,
        cache=cache,
        validate=validate,
        **kwargs,
    )

    # Create dataset
    dataset = tf.data.Dataset.from_generator(
        image_extractor, output_signature=image_extractor.output_signature
    )

    # Assert cardinality
    dataset = dataset.apply(tf.data.experimental.assert_cardinality(len(image_extractor)))

    # Shard dataset
    if num_shards > 1:
        dataset = dataset.shard(num_shards, shard_index)

    # Define helper function to apply map function to dataset
    def dataset_map(dataset, func):
        """Does not apply func to filename."""
        if return_filename:
            return dataset.map(
                lambda x, filename: (func(x), filename),
                num_parallel_calls=tf.data.AUTOTUNE,
            )
        else:
            return dataset.map(func, num_parallel_calls=tf.data.AUTOTUNE)

    # add channel dim
    if len(image_extractor.shape) != 3:
        dataset = dataset_map(dataset, lambda x: tf.expand_dims(x, axis=-1))

    # Clip to image range
    if clip_image_range and image_range is not None:
        dataset = dataset_map(
            dataset,
            partial(
                tf.clip_by_value,
                clip_value_min=image_range[0],
                clip_value_max=image_range[1],
            ),
        )

    # Check if there are outliers in the image range
    if assert_image_range and image_range is not None:
        dataset = dataset_map(dataset, partial(_assert_image_range, image_range=image_range))

    if image_size or resize_type:
        if frame_axis != -1:
            assert resize_axes is not None, (
                "Resizing only works with frame_axis = -1. Alternatively, "
                "you can specify resize_axes."
            )

        # Let resizer handle the assertions.
        resizer = Resizer(
            image_size=image_size,
            resize_type=resize_type,
            resize_axes=resize_axes,
            seed=seed,
            **resize_kwargs,
        )
        dataset = dataset_map(dataset, resizer)

    # repeat dataset if needed (used for smaller datasets)
    if dataset_repetitions:
        dataset = dataset.repeat(dataset_repetitions)

    # batch
    if batch_size:
        dataset = dataset.batch(batch_size, drop_remainder=drop_remainder)

    # normalize
    if normalization_range is not None:
        dataset = dataset_map(
            dataset,
            lambda x: translate(x, image_range, normalization_range),
        )

    # augmentation
    if augmentation is not None:
        dataset = dataset_map(dataset, augmentation)

    # prefetch
    if prefetch:
        dataset = dataset.prefetch(tf.data.AUTOTUNE)

    if wrap_in_keras:
        dataset = TFDatasetToKeras(dataset)

    return dataset



"""Interactive selection tools.

This module provides interactive tools for selecting regions of interest (ROIs)
from 2D arrays or images displayed with matplotlib. It is designed for use in
ultrasound and image processing workflows where manual or semi-automatic selection
of regions is required.

Key Features
------------
- Interactive selection using rectangle or lasso tools via matplotlib widgets.
- Support for cropping, masking, and extracting selected regions from images.
- Polygon and rectangle extraction, interpolation, and mask reconstruction.
- Utilities for batch selection, mask interpolation across frames, and animation.
- Integration with tkinter dialogs for user-friendly selection and confirmation.
- Metric computation (e.g., GCNR) on selected patches.


Example
-------
.. code-block:: python

    import matplotlib.pyplot as plt
    from zea.tools.selection_tool import interactive_selector

    image = ...  # Load your 2D image array
    fig, ax = plt.subplots()
    ax.imshow(image, cmap="gray")
    patches, masks = interactive_selector(image, ax, selector="rectangle")

"""

import tkinter as tk
from collections.abc import Iterable
from pathlib import Path
from typing import Union

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.animation import FuncAnimation
from matplotlib.patches import PathPatch, Rectangle
from matplotlib.path import Path as pltPath
from matplotlib.widgets import LassoSelector, RectangleSelector
from PIL import Image, ImageDraw
from scipy.interpolate import interp1d
from skimage import measure
from skimage.measure import approximate_polygon, find_contours
from sklearn.metrics import pairwise_distances

from zea import log
from zea.internal.viewer import (
    filename_from_window_dialog,
    get_matplotlib_figure_props,
    move_matplotlib_figure,
)
from zea.io_lib import _SUPPORTED_VID_TYPES, load_image, load_video
from zea.metrics import get_metric
from zea.utils import translate


def crop_array(array, value=None):
    """Crop an array to remove all rows and columns containing only a given value."""
    array = np.array(array)
    assert array.ndim == 2, f"Array must be 2D, not {array.ndim}D."
    mask = np.all(np.equal(array, value), axis=1)
    array = array[~mask]

    mask = np.all(np.equal(array, value), axis=0)
    array = array[:, ~mask]
    return array


def interactive_selector(
    data,
    ax,
    selector: str = "rectangle",
    extent: list = None,
    verbose: bool = True,
    num_selections: int = None,
    confirm_selection: bool = True,
) -> tuple:
    """Interactively select part of an array displayed as an image with matplotlib.

    Args:
        data (ndarray): input array. should be 2D.
        ax (plt.ax): existing matplotlib figure ax to select region on.
        selector (str, optional): type of selector. Defaults to 'rectangle'.
            For `lasso` use `LassoSelector`; for `rectangle`, use `RectangleSelector`.
        extent (list): extent of axis where selection is made. Used to transform
            coordinates back to pixel values. Defaults to None.
        verbose (bool): verbosity of print statements. Defaults to False.
        num_selections (int): number of selections to make. Defaults to None.
        confirm_selection (bool): whether to confirm selection before moving on.
            Defaults to True.

    Returns:
        patches (list): list of selected parts of data
        masks (list): list of boolean masks for selected parts of data
    """
    assert data.ndim == 2, f"Data must be 2D, not {data.ndim}D."

    x, y = np.meshgrid(np.arange(data.shape[1], dtype=int), np.arange(data.shape[0], dtype=int))
    pix = np.vstack((x.flatten(), y.flatten())).T

    def _translate_coordinates(x, y):
        if extent:
            x = translate(x, (extent[0], extent[1]), (0, data.shape[1]))
            y = translate(y, (extent[2], extent[3]), (0, data.shape[0]))
        return x, y

    def _onselect_lasso(verts):
        nonlocal select_idx
        if verbose:
            print(f"Selection {select_idx} done")
        select_idx += 1
        verts = np.array(verts)
        # if axis is drawn with extent argument, first translate coordinates to pixels
        verts = np.array(_translate_coordinates(*verts.T)).T
        p = pltPath(verts)
        ind = p.contains_points(pix, radius=1)
        mask.flat[ind] = True
        masks.append(np.copy(mask))
        mask.flat[ind] = False

    def _onselect_rectangle(start, end):
        nonlocal select_idx
        if verbose:
            print(f"Selection {select_idx} done")
        select_idx += 1
        # if axis is drawn with extent argument, first translate coordinates to pixels
        start.xdata, start.ydata = _translate_coordinates(start.xdata, start.ydata)
        end.xdata, end.ydata = _translate_coordinates(end.xdata, end.ydata)

        verts = np.array(
            [
                [start.xdata, start.ydata],
                [start.xdata, end.ydata],
                [end.xdata, end.ydata],
                [end.xdata, start.ydata],
            ],
            int,
        )
        p = pltPath(verts)
        ind = p.contains_points(pix, radius=1)
        mask.flat[ind] = True
        masks.append(np.copy(mask))
        mask.flat[ind] = False

    name_to_selector = {"lasso": LassoSelector, "rectangle": RectangleSelector}
    selector = name_to_selector[selector]
    onselect_dict = {
        LassoSelector: _onselect_lasso,
        RectangleSelector: _onselect_rectangle,
    }
    kwargs_dict = {LassoSelector: {}, RectangleSelector: {"interactive": True}}

    def _execute_selector():
        lasso = selector(ax, onselect_dict[selector], **kwargs_dict[selector])

        if num_selections:
            if verbose:
                print(f"...Plot will close after {num_selections} selections...")
            plt.show(block=False)
            while not select_idx >= num_selections:
                plt.pause(0.1)
        else:
            plt.show(block=False)
            input("Press Enter to continue (don't close plot)...\n")

        lasso.disconnect_events()
        lasso.set_visible(False)
        lasso.update()

    mask = np.tile(False, data.shape)
    masks = []
    select_idx = 0
    _execute_selector()

    patches = []
    for mask in masks:
        patches.append(crop_array(data * mask, value=0))

    like_selection = not bool(confirm_selection)

    while not like_selection:
        print(f"You have made {len(patches)} selection(s).")
        # draw masks on top of data
        for mask in masks:
            add_shape_from_mask(ax, mask, alpha=0.5)
        plt.draw()
        # tkinter yes / no dialog
        like_selection = tk.messagebox.askyesno("Like Selection", "Do you like your selection?")

        if not like_selection:
            remove_masks_from_axs(ax)
            mask = np.tile(False, data.shape)
            masks = []
            select_idx = 0
            _execute_selector()

            patches = []
            for mask in masks:
                patches.append(crop_array(data * mask, value=0))

    return patches, masks


def add_rectangle_from_mask(ax, mask, **kwargs):
    """add a rectangle box to axis from mask array.

    Args:
        ax (plt.ax): matplotlib axis
        mask (ndarray): numpy array with rectangle non-zero
            box defining the region of interest.
    Kwargs:
        edgecolor (str): color of the shape's edge
        facecolor (str): color of the shape's face
        linewidth (int): width of the shape's edge

    Returns:
        plt.ax: matplotlib axis with rectangle added
    """
    # Create a Rectangle patch
    y1, y2 = np.where(np.diff(mask, axis=0).sum(axis=1))[0]
    x1, x2 = np.where(np.diff(mask, axis=1).sum(axis=0))[0]
    rect = Rectangle(
        (x1, y1),
        (x2 - x1),
        (y2 - y1),
        **kwargs,
    )

    # Add the patch to the Axes
    rect_obj = ax.add_patch(rect)
    return rect_obj


def add_shape_from_mask(ax, mask, **kwargs):
    """add a shape to axis from mask array.

    Args:
        ax (plt.ax): matplotlib axis
        mask (ndarray): numpy array with non-zero
            shape defining the region of interest.
    Kwargs:
        edgecolor (str): color of the shape's edge
        facecolor (str): color of the shape's face
        linewidth (int): width of the shape's edge

    Returns:
        plt.ax: matplotlib axis with shape added
    """
    # Create a Path patch
    contours = measure.find_contours(mask, 0.5)
    patches = []
    for contour in contours:
        path = pltPath(contour[:, ::-1])
        patch = PathPatch(path, **kwargs)
        patches.append(ax.add_patch(patch))
    return patches


def interactive_selector_with_plot_and_metric(
    data,
    ax=None,
    selector="rectangle",
    metric=None,
    cmap="gray",
    plot=True,
    mask_plot=False,
    selection_axis=0,
    **kwargs,
):
    """Wrapper for interactive_selector to plot the selected regions.

    Args:
        data (ndarray or list of ndarray): input data.
        ax (plt.ax or list of plt.ax, optional): axis corresponding to input data.
            Defaults to None. In that case function plots data first to create axis.
        selector (str, optional): type of selection tool. Defaults to 'rectangle'.
        metric (str, optional): metric to compute. Defaults to None.
        cmap (str, optional): color map to display data in. Defaults to 'gray'.
        plot (bool, optional): whether to plot selections / metrics on top of axis.
            Defaults to True.
        mask_plot (bool, optional): whether to also plot the masks in a separate plot.
            Can be useful to isolate the patches and see the selections more clearly.
            Defaults to False.
        selection_axis (int, optional): axis on which to make selection. Defaults to 0.

    Raises:
        ValueError: Can only select two patches to compute metric with. More patches
            don't make sense in this context.
    """
    if not isinstance(data, list):
        data = [data]

    if ax is None:
        fig, ax = plt.subplots(1, len(data))
        for _data, _ax in zip(data, ax):
            _ax.imshow(_data, cmap=cmap, aspect="auto")

    if not isinstance(ax, Iterable):
        ax = [ax]

    # create selector for first axis only
    patches, masks = interactive_selector(
        data[selection_axis], ax[selection_axis], selector, num_selections=2, **kwargs
    )

    if len(patches) != 2:
        raise ValueError("exactly 2 patches are required for using this wrapper function")

    # get patches for all data in data list using the selection made
    patches = []
    for image in data:
        patches.extend([crop_array(image * mask, value=0) for mask in masks])

    # compute metrics
    scores = []
    if metric:
        for i in range(len(data)):
            idx = i * len(masks)
            score = get_metric(metric)(patches[idx], patches[idx + 1])
            scores.append(score)
            print(f"{metric}: {score:.3f}")

    # plot on top of existing plot
    if plot:
        for _ax, score in zip(ax, scores):
            title = _ax.get_title()
            _ax.set_title(title + "\n" + f"{metric}: {score:.3f}")
            for mask in masks:
                if selector == "rectangle":
                    add_rectangle_from_mask(_ax, mask, alpha=0.5)
                else:
                    add_shape_from_mask(_ax, mask, alpha=0.5)
            plt.tight_layout()

    # plot patches and masks
    if mask_plot:
        fig, axs = plt.subplots(len(masks), 3)
        for i, (ax_new, patch, mask) in enumerate(zip(axs, patches, masks)):
            if i == 0:
                ax_base = ax_new[selection_axis]
                ax_base.imshow(data[selection_axis], cmap=cmap, aspect="auto")
            ax_new[1].imshow(patch, cmap=cmap, aspect="auto")
            ax_new[2].imshow(mask, aspect="auto")

            if selector == "rectangle":
                add_rectangle_from_mask(ax_base, mask)

            for _ax in ax_new:
                _ax.axis("off")

        fig.tight_layout()

    return scores


def extract_rectangle_from_mask(image):
    """Find corner points of rectangle in binary mask.
    Args:
        image (np.ndarray): 2D binary mask
    Returns:
        Tuple of the form ((x1, y1), (x2, y2)) with the corner points of the rectangle.
    """
    image = np.array(image)
    indices = np.argwhere(image == 1)
    if len(indices) == 0:
        return None
    top, left = indices.min(axis=0)
    bottom, right = indices.max(axis=0)
    return ((left, top), (right, bottom))


def reconstruct_mask_from_rectangle(corner_points, image_shape):
    """Reconstruct a binary mask from corner points of a rectangle.

    Args:
        corner_points (tuple): Tuple of the form ``((x1, y1), (x2, y2))``
            with the corner points of the rectangle.
        image_shape (tuple): Size of the image (height, width).

    Returns:
        np.ndarray: 2D boolean mask of shape (height, width).

    """
    image = np.zeros(image_shape, dtype=bool)
    x1, y1 = corner_points[0]
    x2, y2 = corner_points[1]
    image[y1 : y2 + 1, x1 : x2 + 1] = True
    return image


def interpolate_rectangles(rectangles, x_indices, y_indices):
    """Interpolate between arbitrary number of rectangles.

    Args:
        rectangles (list): List with any number of rectangles as tuples of the form
            ((x1, y1), (x2, y2)). Size of the list must be equal to the number of x indices.
        x_indices (np.ndarray): Array with x indices for interpolation.
        y_indices (np.ndarray): Array with y indices for interpolation.

    Returns:
        List with interpolated rectangles as tuples of the form ((x1, y1), (x2, y2)).
            Size of the list is equal to the number of y indices.
    """
    new_rectangles = []
    x1 = [rect[0][0] for rect in rectangles]
    x2 = [rect[1][0] for rect in rectangles]
    y1 = [rect[0][1] for rect in rectangles]
    y2 = [rect[1][1] for rect in rectangles]

    values_interp = []
    for values in [x1, x2, y1, y2]:
        values_interp.append(np.interp(y_indices, x_indices, values).astype(np.int32))

    x1, x2, y1, y2 = values_interp
    new_rectangles = [((x1[i], y1[i]), (x2[i], y2[i])) for i in range(len(x1))]
    return new_rectangles


def extract_polygon_from_mask(mask, tolerance: float = 0.01, verbose: bool = True):
    """Find largest contour in a binary mask and fit polygon.

    Polygon approximation will reduce contour points, unless tolerance is 0.

    Args:
        mask (np.ndarray): 2D binary mask
        tolerance (float): Approximation tolerance for polygonal contour
    Returns:
        Numpy array of shape (N, 2) with vertices of the polygon.
    """
    contours = find_contours(mask, 0.5, fully_connected="high")
    # return the largest contour
    if len(contours) > 1:
        contour_lengths = [len(contour) for contour in contours]
        contour = contours[np.argmax(contour_lengths)]
        if verbose:
            log.warning("Warning: multiple contours found. Returning the largest contour.")
    elif len(contours) == 0:
        if verbose:
            log.warning("Warning: no contours found. Returning None.")
        return None
    else:
        contour = contours[0]
    poly = approximate_polygon(contour, tolerance)
    return poly


def reconstruct_mask_from_polygon(vertices, image_size):
    """Reconstruct a binary mask from a polygon.

    Fills in regions defined by the polygon contour.
    Args:
        vertices (np.ndarray): Vertices of the polygon as an array of shape (N, 2).
        image_size (tuple): Size of the image (height, width).
    Returns:
        np.ndarray (height, width) with the reconstructed mask.
    """
    # Create a path for the polygon
    mask = Image.new("L", (image_size[1], image_size[0]), 0)

    # Create a draw object
    draw = ImageDraw.Draw(mask)

    # Close the polygon by adding the first point to the end
    vertices = np.vstack((vertices, vertices[0]))

    # Draw the filled polygon on the mask
    polygon_coords = [(x, y) for y, x in vertices]
    draw.polygon(polygon_coords, outline=1, fill=1)

    # Convert the mask to a NumPy array
    mask_array = np.array(mask)
    return mask_array


def interpolate_polygons(polygon1, polygon2, t):
    """Interpolate between two polygons.
    Args:
        polygon1 (np.ndarray): First polygon as an array of shape (N, 2).
        polygon2 (np.ndarray): Second polygon as an array of shape (N, 2).
        t (float): Interpolation parameter, where 0 <= t <= 1.
    Returns:
        Interpolated polygon as an array of shape (N, 2).
    """
    # Ensure both polygons have the same number of vertices
    if polygon1.shape[0] != polygon2.shape[0]:
        raise ValueError("Both polygons must have the same number of vertices.")

    # Perform linear interpolation for each vertex
    interpolated_polygon = (1 - t) * polygon1 + t * polygon2

    return interpolated_polygon


def match_polygons(polygon1, polygon2):
    """Match two polygons by minimizing the total distance between vertices.

    The vertices of the first polygon are shifted circularly to find the best match.
    Order of vertices is preserved.

    Args:
        polygon1 (np.ndarray): First polygon as an array of shape (N, 2).
        polygon2 (np.ndarray): Second polygon as an array of shape (N, 2).
    Returns:
        Tuple of the form (poly1, poly2), where poly1 and poly2 are the matched polygons.
    """

    distances = pairwise_distances(polygon1, polygon2, metric="euclidean")

    min_total_distance = float("inf")
    best_shift = 0

    # Find the shift that minimizes the total distance.
    n, m = distances.shape
    for shift in range(n):
        total_distance = 0
        for i in range(n):
            total_distance += distances[i, (i + shift) % m]
        if total_distance < min_total_distance:
            min_total_distance = total_distance
            best_shift = shift

    polygon1 = np.roll(polygon1, best_shift, axis=0)
    return polygon1, polygon2


def equalize_polygons(polygons, mode="max"):
    """Make sure all polygons have the same number of vertices.

    Args:
        polygons (list): List with any number of polygons as arrays of shape (N, 2).
        mode (str): Method for equalizing the number of vertices. Either 'max' or 'min'.
            with 'max' the number of vertices is equal to the polygon with the most vertices.
            with 'min' the number of vertices is equal to the polygon with the least vertices.
    Returns:
        A tuple of the form (poly1, poly2, ...), where poly1, poly2, ...
            are the trimmed polygons with the same number of vertices as the
            polygon with the fewest / most vertices, depending on the mode.
    """
    assert mode in ["max", "min"], f"Mode must be either 'max' or 'min', not {mode}."
    if mode == "max":
        num_vertices = max(polygon.shape[0] for polygon in polygons)
    elif mode == "min":
        num_vertices = min(polygon.shape[0] for polygon in polygons)
    else:
        raise ValueError(f"Mode must be either 'max' or 'min', not {mode}.")

    # give warning if difference in min / max vertices is large
    if num_vertices < 0.8 * max(polygon.shape[0] for polygon in polygons):
        log.warning(
            "Warning: difference in number of vertices is large. "
            "Possibly due to large difference in polygon size."
        )

    if mode == "min":
        trimmed_polygons = []
        for polygon in polygons:
            indices = np.linspace(0, len(polygon) - 1, num_vertices).astype(int)
            trimmed_polygons.append(polygon[indices])

        return trimmed_polygons
    elif mode == "max":
        # interpolate the contours
        interpolated_polygons = []
        for polygon in polygons:
            if polygon.shape[0] < num_vertices:
                # interp2d
                indices = np.linspace(0, len(polygon) - 1, num_vertices)

                # create a function to interpolate the x and y coordinates separately
                f_x = interp1d(np.arange(len(polygon)), polygon[:, 0], kind="linear")
                f_y = interp1d(np.arange(len(polygon)), polygon[:, 1], kind="linear")

                # evaluate the functions at the interpolated indices
                interpolated_polygons.append(np.column_stack((f_x(indices), f_y(indices))))
            else:
                interpolated_polygons.append(polygon)
        return interpolated_polygons


def interpolate_masks(
    masks: Union[list, np.ndarray], num_frames: int, rectangle: bool = False
) -> list:
    """Interpolate between arbitrary number of masks."""
    assert isinstance(masks, (list, np.ndarray)), "Masks must be a list of numpy arrays."
    assert num_frames > 1, "At least two frames are required for interpolation."
    number_of_masks = len(masks)
    assert number_of_masks > 1, "At least two masks are required for interpolation."
    mask_shape = masks[0].shape
    assert all(mask.shape == mask_shape for mask in masks), "All masks must have the same shape."

    # distribute number of frames over number of masks
    num_frames_per_segment = [num_frames // (number_of_masks - 1)] * (number_of_masks - 1)
    if num_frames % num_frames_per_segment[0] != 0:
        # make sure that number of frames per mask adds up to total number of frames
        num_frames_per_segment[-1] += num_frames - sum(num_frames_per_segment)

    if rectangle:
        # get the rectangles
        rectangles = []
        for mask in masks:
            rectangles.append(extract_rectangle_from_mask(mask))

        rectangles = interpolate_rectangles(
            rectangles,
            np.linspace(0, num_frames - 1, len(rectangles)),
            np.arange(num_frames),
        )

        # reconstruct the masks
        interpolated_masks = []
        for _rectangle in rectangles:
            interpolated_masks.append(reconstruct_mask_from_rectangle(_rectangle, mask_shape))
        return interpolated_masks

    # get the contours
    polygons = []
    for mask in masks:
        polygons.append(extract_polygon_from_mask(mask))

    # trim the polygons for equal number of vertices
    polygons = equalize_polygons(polygons)

    # match the polygons
    for i in range(number_of_masks - 1):
        polygons[i], polygons[i + 1] = match_polygons(polygons[i], polygons[i + 1])

    # interpolate the polygons
    interpolated_polygons = []
    for i in range(number_of_masks - 1):
        for t in np.linspace(0, 1, num_frames_per_segment[i]):
            interpolated_polygons.append(interpolate_polygons(polygons[i], polygons[i + 1], t))

    # reconstruct the masks
    interpolated_masks = []
    for interpolated_polygon in interpolated_polygons:
        interpolated_masks.append(reconstruct_mask_from_polygon(interpolated_polygon, mask_shape))

    return interpolated_masks


def interactive_selector_for_dataset():
    """To be added. UI for generating and saving masks for entire dataset.
    In an efficient and user friendly way.
    """
    raise NotImplementedError


def ask_for_selection_tool():
    """Ask user for which selection tool to use."""
    while True:
        selector = input("Which selection tool do you want to use? [rectangle/lasso]): ")
        if selector in ["rectangle", "lasso"]:
            break
        print("Please enter either 'rectangle' or 'lasso'")
    return selector


def ask_for_num_selections():
    """Ask user for number of selections to make."""
    while True:
        num_selections = input("How many selections do you want to make? ")
        try:
            num_selections = int(num_selections)
            if num_selections < 1:
                raise ValueError
            break
        except ValueError:
            print("Please enter a positive integer")
    return num_selections


def ask_save_animation_with_fps():
    """Ask user for fps to save animation with."""
    while True:
        try:
            fps = int(input("Save animation as gif? Enter fps: "))
            break
        except ValueError:
            print("Please enter a positive integer")
    return fps


def remove_masks_from_axs(axs: matplotlib.axes.Axes) -> None:
    """Remove all masks from the given axes object."""
    for obj in axs.findobj():
        if isinstance(obj, (PathPatch, Rectangle)):
            try:
                obj.remove()
            except Exception:
                pass


def update_imshow_with_mask(
    frame_no: int,
    axs: matplotlib.axes.Axes,
    imshow_obj: matplotlib.image.AxesImage,
    images: np.ndarray,
    masks: np.ndarray,
    selector: str,
) -> tuple:
    """Updates the imshow object with the image from the given frame and
    overlays a mask on top of it.

    Args:
        frame_no (int): The index of the frame to display.
        axs (matplotlib.axes.Axes): The axes object to display the image on.
        imshow_obj (matplotlib.image.AxesImage): The imshow object to update.
        images (numpy.ndarray): An array of images to display.
        masks (numpy.ndarray): An array of masks to overlay on top of the images.
        selector (str): The type of selector to use for the mask.
            Can be either "rectangle" or "shape".

    Returns:
        tuple: A tuple containing the updated imshow object and the mask object.
    """
    imshow_obj.set_array(images[frame_no])
    remove_masks_from_axs(axs)
    if selector == "rectangle":
        mask_obj = add_rectangle_from_mask(axs, masks[frame_no])
    else:
        mask_obj = add_shape_from_mask(axs, masks[frame_no], alpha=0.5)
    return imshow_obj, mask_obj


def main():
    """Main function for interactive selector on multiple images."""
    print(
        "Select as many images as you like, OR select 1 video / gif / dicom, "
        "and close window to continue..."
    )
    images = []
    file_names = []
    try:
        while True:
            file = filename_from_window_dialog("Choose image / video file")
            if file.suffix in [".png", ".jpg", ".jpeg"]:
                image = load_image(file)
                images.append(image)
                file_names.append(file.name)
                same_images = True
            elif file.suffix in _SUPPORTED_VID_TYPES:
                images.extend(load_video(file))
                same_images = False
                break
    except Exception as e:
        if len(images) == 0:
            raise e
        print("No more images selected. Continuing...")

    selector = ask_for_selection_tool()

    if same_images is True:
        figs, axs = [], []
        for i, (image, file_name) in enumerate(zip(images[::-1], file_names[::-1])):
            fig, ax = plt.subplots()
            ax.imshow(image, cmap="gray")
            if i == len(images) - 1:
                ax.set_title(f"Make selection in this plot\n {file_name}")
            else:
                ax.set_title(file_name)
            ax.axis("off")
            axs.append(ax)
            figs.append(fig)

        axs = axs[::-1]
        figs = figs[::-1]

        interactive_selector_with_plot_and_metric(
            images,
            axs,
            selector=selector,
            metric="gcnr",
        )

    else:
        if len(images) > 3:
            print(f"Found sequence of {len(images)} images. ")

            num_selections = ask_for_num_selections()

            selection_idx = np.linspace(0, len(images) - 1, int(num_selections)).astype(int)
            selection_images = [images[idx] for idx in selection_idx]
            selection_masks = []
            pos, size = None, None
            for image in selection_images:
                fig, axs = plt.subplots()
                fig.tight_layout()
                # set window size to what user selected for plot before
                if pos is not None:
                    move_matplotlib_figure(fig, pos, size)

                axs.imshow(image, cmap="gray")

                while True:
                    _, mask = interactive_selector(image, axs, selector=selector, num_selections=1)
                    # check if mask is empty else retry
                    if mask[0].sum() == 0:
                        print("Empty mask. Try again, make sure to make a descent selection...")
                    else:
                        break

                pos, size = get_matplotlib_figure_props(fig)

                if selector == "rectangle":
                    add_rectangle_from_mask(axs, mask[0], alpha=0.5)
                else:
                    add_shape_from_mask(axs, mask[0], alpha=0.5)
                plt.close()
                selection_masks.append(mask[0])

        # small hack to make sure that there is always at least two masks for interpolation
        if len(selection_masks) == 1:
            selection_masks.append(selection_masks[0])

        interpolated_masks = interpolate_masks(
            selection_masks, num_frames=len(images), rectangle=(selector == "rectangle")
        )

        fig, axs = plt.subplots()

        imshow_obj = axs.imshow(images[0], cmap="gray")

        if selector == "rectangle":
            add_rectangle_from_mask(axs, interpolated_masks[0])
        else:
            add_shape_from_mask(axs, interpolated_masks[0], alpha=0.5)

        fps = ask_save_animation_with_fps()

        ani = FuncAnimation(
            fig,
            update_imshow_with_mask,
            frames=len(images),
            fargs=(axs, imshow_obj, images, interpolated_masks, selector),
            interval=1000 / fps,
        )
        filename = Path(file.parent.stem + "_" + f"{file.stem}_interpolated_masks.gif")
        ani.save(filename, writer="pillow")
        print(f"Succesfully saved animation as {filename}")


if __name__ == "__main__":
    main()



"""Functions for displaying images and videos using OpenCV or Matplotlib."""

import abc
import os
import sys
from collections import deque
from multiprocessing.pool import ThreadPool
from pathlib import Path
from tkinter import Tk
from tkinter.filedialog import askopenfilename
from typing import Callable, Optional

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.axes_grid1 import make_axes_locatable

from zea import log


def running_in_notebook():
    """Check whether code is running in a Jupyter Notebook or not."""
    return "ipykernel_launcher" in sys.argv[0]


def plt_window_has_been_closed(fig):
    """Checks whether matplotlib plot window is closed"""
    return not plt.fignum_exists(fig.number)


def filename_from_window_dialog(window_name=None, filetypes=None, initialdir=None):
    """Get filename through dialog window
    Args:
        window_name: string with name of window
        filetypes: tuple of tuples containing (name, filetypes)
            example:
                (('mat or hdf5 or whatever you want', '*.mat *.hdf5 *'), (ckpt, *.ckpt))
        initialdir: path to directory where window will start
    Returns:
        filename: string containing path to selected file
    """
    if filetypes is None:
        filetypes = (("all files", "*.*"),)

    try:
        root = Tk()
    except Exception as error:
        raise ValueError(
            "Cannot run zea GUI on a server, unless a X11 server is properly setup"
        ) from error

    # open in foreground
    root.wm_attributes("-topmost", True)
    root.wm_attributes("-topmost", False)

    # we don't want a full GUI, so keep the root window from appearing
    if not running_in_notebook():
        root.withdraw()

    # show an "Open" dialog box and return the path to the selected file
    filename = askopenfilename(
        parent=root,
        title=window_name,
        filetypes=filetypes,
        initialdir=initialdir,
    )
    root.destroy()

    # check whether a file was selected
    if filename:
        return Path(filename)
    else:
        raise ValueError("No file selected.")


def move_matplotlib_figure(figure, position, size=None):
    """Move matplotlib figure to a specific position on the screen.
    Args:
        figure (plt.figure): matplotlib figure
        position (tuple): x and y position of figure in pixels
        size (tuple, optional): width and height of figure in pixels

    """
    x, y = position

    if size is not None:
        width, height = size
        figure.set_size_inches(width / figure.dpi, height / figure.dpi)

    backend = matplotlib.get_backend()

    if backend == "TkAgg":
        figure.canvas.manager.window.wm_geometry(f"+{x}+{y}")
    elif backend == "WXAgg":
        figure.canvas.manager.window.SetPosition((x, y))
    else:
        # This works for QT and GTK
        # You can also use window.setGeometry
        figure.canvas.manager.window.move(x, y)


def get_matplotlib_figure_props(figure):
    """Return a dictionary of matplotlib figure properties.
    Args:
        figure (plt.figure): matplotlib figure
    Returns:
        tuple: position and size of figure in pixels
            position (tuple): x and y position of figure in pixels
            size (tuple): width and height of figure in pixels
    """
    # Get size and position in the format of "widthxheight+X+Y"
    geometry = figure.canvas.manager.window.geometry()

    # Get the geometry object based on the backend
    backend_name = matplotlib.get_backend()

    try:
        if backend_name == "TkAgg":
            assert isinstance(geometry, str), (
                f"Unsupported geometry type: {type(geometry)} for backend: {backend_name}"
            )
            # format: "widthxheight+X+Y"
            # Split the geometry string by '+' to extract size and position
            size_str, *pos_str = geometry.split("+")
            # Extract width and height from the size string
            size = map(int, size_str.split("x"))
            # Extract X and Y position values as integers
            position = map(int, pos_str)
        elif backend_name == "QtAgg":
            # format: QRect object
            position = geometry.x(), geometry.y()
            size = geometry.size().width(), geometry.size().height()
        else:
            raise ValueError(f"Unsupported backend: {backend_name}")
    except Exception as error:
        log.warning(f"Could not get figure properties: {error}")
        position, size = None, None

    return position, size


def raise_matplotlib_window(figname=None):
    """Raise the matplotlib window for Figure figname to the foreground.

    If no argument is given, raise the current figure.

    This function will only work with a Qt or Tk graphics backend.
    """
    # check for backend
    backend = matplotlib.get_backend()
    if backend in ["Qt5Agg", "TkAgg"]:
        log.warning("This function only works with a Qt or Tk graphics backend")

    if figname:
        plt.figure(figname)

    cfm = plt.get_current_fig_manager()

    if backend in ["QtAgg", "Qt4Agg", "Qt5Agg"]:
        cfm.window.activateWindow()
        cfm.window.raise_()
    elif backend == "TkAgg":
        cfm.window.attributes("-topmost", True)
        cfm.window.attributes("-topmost", False)


class DummyTask:
    """Dummy task that returns the data passed to it.

    Used in ImageViewer to pass data to the ImageViewer without using threads.
    """

    def __init__(self, data):
        self.data = data

    def ready(self):
        """Always returns True."""
        return True

    def get(self):
        """Gets the data passed to the DummyTask."""
        return self.data


class ImageViewer(abc.ABC):
    """ImageViewer is an abstract class for displaying frames in a non-blocking way."""

    def __init__(
        self,
        get_frame: Callable[[], np.ndarray],
        window_name: Optional[str] = "frame",
        num_threads: Optional[int] = None,
        resizable_window: Optional[bool] = True,
        threading: Optional[bool] = True,
        collect_frames: Optional[bool] = False,
    ) -> None:
        """Initializes the ImageViewer object.
        Args:
            get_frame (function): A function that returns the current frame.
            window_name (str): The name of the window to display the frame.
            num_threads (int, optional): The number of threads to use for processing frames.
            resizable_window (bool, optional): Whether the window should be resizable or not.
            threading (bool, optional): Whether to use threading or not.
            collect_frames (bool, optional): Whether to collect frames or not.
        """
        self.get_frame_func = get_frame
        self.window_name = window_name
        self.num_threads = num_threads
        self.resizable_window = resizable_window
        self.threading = threading
        self.collect_frames = collect_frames

        if self.num_threads is None:
            # Default to the number of CPU cores if not specified
            self.num_threads = max(1, os.cpu_count() - 1)

        if self.threading:
            self.pool = ThreadPool(processes=self.num_threads)

        self.pending = deque()
        self.frame_no = 0
        self.frames = []

    @abc.abstractmethod
    def show(self, *args, **kwargs) -> None:
        """Displays a frame.
        Frame is generated using the get_frame function passed during initialization.

        This function is non-blocking, and will return immediately.
        """
        self._add_task(*args, **kwargs)
        # show the frame

    def _add_task(self, *args, **kwargs):
        if self.threading:
            if len(self.pending) < self.num_threads:
                task = self.pool.apply_async(self.get_frame_func, args=args, kwds=kwargs)
                self.pending.append(task)
        else:
            task = DummyTask(self.get_frame_func(*args, **kwargs))
            self.pending.append(task)

    def _get_frame(self):
        """Returns the most recent frame in the queue."""
        if len(self.pending) == 0:
            raise ValueError("No frames available.")
        if not self.pending[0].ready():
            raise ValueError("Frame not ready.")
        frame = self.pending.popleft().get()
        if self.collect_frames:
            self.frames.append(frame)
        return frame

    @property
    def frame_is_ready(self):
        """Returns True if a frame is ready to be displayed."""
        return len(self.pending) > 0 and self.pending[0].ready()


class ImageViewerOpenCV(ImageViewer):
    """ImageViewer displays frames using OpenCV's imshow function in a non-blocking way.

    Example:
        >>> import numpy as np
        >>> from zea.internal.io_lib import ImageViewerOpenCV
        >>> def generate_frame():
        >>>     return np.random.randint(0, 255, (400, 600, 3), dtype=np.uint8)
        >>> image_viewer = ImageViewerOpenCV(generate_frame, threading=True, num_threads=1)
        >>> while True:
        >>>     image_viewer.show()
        >>>     if cv2.waitKey(25) & 0xFF == ord("q"):
        >>>         break
    """

    def __init__(
        self,
        get_frame: Callable[[], np.ndarray],
        window_name: Optional[str] = "frame",
        num_threads: Optional[int] = None,
        resizable_window: Optional[bool] = True,
        threading: Optional[bool] = True,
        headless: Optional[bool] = False,
    ) -> None:
        """Initializes the ImageViewerOpenCV object."""
        super().__init__(get_frame, window_name, num_threads, resizable_window, threading)
        self.window = None
        self.headless = headless

        try:
            import cv2

            self._cv2 = cv2
        except ImportError as exc:
            raise ImportError(
                "OpenCV is required for ImageViewerOpenCV to work. "
                "Please install it with 'pip install opencv-python' or "
                "'pip install opencv-python-headless'."
            ) from exc

    def _create_window(self):
        if self.resizable_window:
            self._cv2.namedWindow(self.window_name, self._cv2.WINDOW_NORMAL)
        else:
            self._cv2.namedWindow(self.window_name)
        self.window = True

    def show(self, *args, **kwargs) -> None:
        """Displays a frame using OpenCV's imshow function.
        Frame is generated using the get_frame function passed during initialization.

        This function is non-blocking, and will return immediately.
        Imshow is called asynchronously in a separate thread.
        """
        super().show(*args, **kwargs)
        while self.frame_is_ready:
            frame = self._get_frame()
            frame = np.array(frame, dtype=np.uint8)
            if self.headless:
                self.frame_no += 1
                continue

            if self.frame_no == 0:
                if self.window is None:
                    self._create_window()
                    self._cv2.resizeWindow(self.window_name, frame.shape[1], frame.shape[0])

            self._cv2.imshow(self.window_name, frame)
            self.frame_no += 1

    def close(self):
        """Closes the window."""
        if not self.window:
            return
        if self.has_been_closed():
            return
        self._cv2.destroyWindow(self.window_name)

    def has_been_closed(self):
        """Returns True if the window has been closed."""
        if self.window is None:
            return False
        return self._cv2.getWindowProperty(self.window_name, self._cv2.WND_PROP_VISIBLE) < 1


class ImageViewerMatplotlib(ImageViewer):
    """ImageViewerMatplotlib displays frames using matplotlib's imshow function
    in a non-blocking way.

    Example:
        >>> import numpy as np
        >>> from zea.internal.io_lib import ImageViewerMatplotlib, plt_window_has_been_closed
        >>> def generate_frame():
        >>>     return np.random.randint(0, 255, (400, 600, 3), dtype=np.uint8)
        >>> image_viewer = ImageViewerMatplotlib(generate_frame, threading=True, num_threads=1)
        >>> while True:
        >>>     image_show.show()
        >>>     if cv2.waitKey(25) & plt_window_has_been_closed(image_viewer.fig):
        >>>         break
    """

    def __init__(
        self,
        get_frame: Callable[[], np.ndarray],
        window_name: Optional[str] = "frame",
        num_threads: Optional[int] = None,
        resizable_window: Optional[bool] = True,
        threading: Optional[bool] = True,
        imshow_kwargs: Optional[dict] = None,
        cax_kwargs: Optional[dict] = None,
    ) -> None:
        """Initializes the ImageViewerMatplotlib object."""
        super().__init__(get_frame, window_name, num_threads, resizable_window, threading)
        plt.ion()
        self.fig = None
        self.ax = None
        self.bg = None
        self.image_obj = None
        self.imshow_kwargs = imshow_kwargs
        self.cax_kwargs = cax_kwargs

        self.init_figure_props = True

    def _create_figure(self):
        # create figure but do not show it already
        self.fig = plt.figure(self.window_name)
        self.ax = self.fig.add_subplot(111)
        self.bg = self.fig.canvas.copy_from_bbox(self.fig.bbox)
        # move to foreground
        raise_matplotlib_window(self.window_name)

        if not self.resizable_window:
            backend = matplotlib.get_backend()
            if backend == "Qt4Agg":
                win = self.fig.canvas.window()
                win.setFixedSize(win.size())
            else:
                log.warning(f"Backend {backend} does not support fixed size windows.")

    def show(self, *args, **kwargs) -> None:
        """Displays a frame using matplotlib's imshow function.
        Frame is generated using the get_frame function passed during initialization.

        This function is non-blocking, and will return immediately.
        """
        super().show(*args, **kwargs)

        while self.frame_is_ready:
            frame = self._get_frame()

            if self.image_obj is None:
                if self.fig is None:
                    self._create_figure()
                if self.bg is None:
                    self.bg = self.fig.canvas.copy_from_bbox(self.fig.bbox)

                if self.imshow_kwargs:
                    self.image_obj = self.ax.imshow(frame, **self.imshow_kwargs, animated=True)
                else:
                    self.image_obj = self.ax.imshow(frame, animated=True)

                # these only need to be set once
                if self.init_figure_props:
                    if self.cax_kwargs:
                        divider = make_axes_locatable(self.ax)
                        cax = divider.append_axes(**self.cax_kwargs)
                        if "color" in self.cax_kwargs:
                            color = self.cax_kwargs.pop("color")
                            cax.yaxis.label.set_color(color)
                            cax.tick_params(axis="y", colors=color)
                            cax.title.set_color(color)
                        plt.colorbar(self.image_obj, cax=cax)
                    self.fig.tight_layout()
                    self.init_figure_props = False

                self.fig.canvas.draw()
            else:
                self.image_obj.set_data(frame)
                self.fig.canvas.blit(self.fig.bbox)
                self.fig.canvas.draw()

            self.fig.canvas.flush_events()
            self.frame_no += 1

    def close(self):
        """Closes the window."""
        plt.close(self.fig)

    def has_been_closed(self):
        """Returns True if the window has been closed."""
        return plt_window_has_been_closed(self.fig)



"""Gaussian Mixture Model (GMM) implementation"""

import keras
import numpy as np
from keras import ops

from zea.models.generative import GenerativeModel
from zea.tensor_ops import linear_sum_assignment


class GaussianMixtureModel(GenerativeModel):
    """
    Gaussian Mixture Model fitted with EM algorithm.

    Args:
        n_components: Number of mixture components.
        n_features: Number of features (dimensions).
        max_iter: Maximum number of EM steps.
        tol: Convergence tolerance.
        seed: Random seed for reproducibility.

    Example:
    ```python
    gmm = GaussianMixtureModel(n_components=2, n_features=2)
    gmm.fit(data, max_iter=100)
    samples = gmm.sample(100)
    ```
    """

    def __init__(self, n_components=2, n_features=1, tol=1e-4, seed=None):
        self.n_components = n_components
        self.n_features = n_features
        self.tol = tol
        self.seed = seed
        self._initialized = False

        self.means = None  # (n_components, n_features)
        self.vars = None  # (n_components, n_features)
        self.pi = None  # (n_components,)

    def _initialize(self, X):
        # X: (n_samples, n_features)
        n_samples = ops.shape(X)[0]
        n_features = ops.shape(X)[1]
        chosen = []
        # Pick the first mean randomly
        idx = ops.cast(
            keras.random.uniform(
                shape=(),
                minval=0,
                maxval=n_samples,
                seed=self.seed,
            ),
            "int32",
        )
        chosen.append(idx)
        for _ in range(1, self.n_components):
            # Gather chosen means so far
            chosen_means = ops.stack(
                [ops.take(X, i, axis=0) for i in chosen], axis=0
            )  # (len(chosen), n_features)
            # Compute distances from all points to each chosen mean
            # (n_samples, len(chosen), n_features)
            diffs = ops.expand_dims(X, 1) - ops.expand_dims(chosen_means, 0)
            dists = ops.sqrt(ops.sum(diffs**2, axis=-1))  # (n_samples, len(chosen))
            min_dists = ops.min(dists, axis=1)  # (n_samples,)
            idx = ops.argmax(min_dists, axis=0)
            chosen.append(idx)
        means = ops.stack([ops.take(X, i, axis=0) for i in chosen], axis=0)
        self.means = means
        # Initialize variances to variance of data
        var = ops.var(X, axis=0)
        self.vars = ops.ones((self.n_components, n_features)) * var
        # Initialize mixture weights uniformly
        self.pi = ops.ones((self.n_components,)) / self.n_components
        self._initialized = True

    def _e_step(self, X):
        # X: (n_samples, n_features)
        X_exp = ops.expand_dims(X, axis=1)  # (n_samples, 1, n_features)
        means = ops.expand_dims(self.means, axis=0)  # (1, n_components, n_features)
        vars_ = ops.expand_dims(self.vars, axis=0)  # (1, n_components, n_features)
        pi = self.pi  # (n_components,)

        # Compute log Gaussian pdf for each component
        log_prob = -0.5 * ops.sum(
            ops.log(2 * np.pi * vars_) + ((X_exp - means) ** 2) / vars_, axis=-1
        )  # (n_samples, n_components)
        # Add log mixture weights
        log_prob = log_prob + ops.log(pi)
        # Normalize to get responsibilities
        log_prob_norm = log_prob - ops.logsumexp(log_prob, axis=1, keepdims=True)
        gamma = ops.exp(log_prob_norm)  # (n_samples, n_components)
        return gamma  # responsibilities

    def _m_step(self, X, gamma):
        # X: (n_samples, n_features)
        # gamma: (n_samples, n_components)
        Nk = ops.sum(gamma, axis=0)  # (n_components,)
        # Update means
        means = ops.sum(
            ops.expand_dims(gamma, -1) * ops.expand_dims(X, 1), axis=0
        ) / ops.expand_dims(Nk, -1)
        # Update variances
        X_exp = ops.expand_dims(X, axis=1)  # (n_samples, 1, n_features)
        means_exp = ops.expand_dims(means, axis=0)  # (1, n_components, n_features)
        vars_ = ops.sum(gamma[..., None] * (X_exp - means_exp) ** 2, axis=0) / ops.expand_dims(
            Nk, -1
        )
        # Update mixture weights
        pi = Nk / ops.sum(Nk)
        return means, vars_, pi

    def fit(self, data, max_iter=100, verbose=0, **kwargs):
        X = ops.convert_to_tensor(data, dtype="float32")
        if not self._initialized:
            self._initialize(X)

        prev_ll = None
        progbar = keras.utils.Progbar(max_iter, verbose=verbose)
        for i in range(max_iter):
            # E-step
            gamma = self._e_step(X)
            # M-step
            means, vars_, pi = self._m_step(X, gamma)
            # Compute log-likelihood
            self.means, self.vars, self.pi = means, vars_, pi
            ll = ops.sum(ops.log(ops.sum(self._component_pdf(X) * self.pi, axis=1)))
            if verbose:
                progbar.update(i + 1, values=[("log-likelihood", float(ll))])
            if prev_ll is not None and abs(float(ll) - float(prev_ll)) < self.tol:
                if verbose:
                    print(f"\nConverged at iter {i}")
                break
            prev_ll = ll

    def _component_pdf(self, X):
        # X: (n_samples, n_features)
        X_exp = ops.expand_dims(X, axis=1)  # (n_samples, 1, n_features)
        means = ops.expand_dims(self.means, axis=0)  # (1, n_components, n_features)
        vars_ = ops.expand_dims(self.vars, axis=0)  # (1, n_components, n_features)
        # Gaussian PDF (no mixture weights)
        norm = ops.prod(ops.sqrt(2 * np.pi * vars_), axis=-1)
        exp_term = ops.exp(-0.5 * ops.sum(((X_exp - means) ** 2) / vars_, axis=-1))
        return exp_term / norm  # (n_samples, n_components)

    def sample(self, n_samples=1, seed=None, **kwargs):
        # Sample component indices
        comp_idx = keras.random.categorical(ops.log(self.pi[None, :]), n_samples, seed=seed)
        comp_idx = ops.squeeze(comp_idx, axis=0)
        means = ops.take(self.means, comp_idx, axis=0)
        vars_ = ops.take(self.vars, comp_idx, axis=0)
        eps = keras.random.normal(ops.shape(means), seed=seed)
        samples = means + eps * ops.sqrt(vars_)
        return samples

    def posterior_sample(self, measurements, n_samples=1, seed=None, **kwargs):
        """
        Sample component indices from the posterior p(z|x) for each measurement.

        Args:
            measurements: Input data, shape (batch, n_features).
            n_samples: Number of posterior samples per measurement.
            seed: Random seed.

        Returns:
            Component indices, shape (batch, n_samples).
        """
        X = ops.convert_to_tensor(measurements, dtype="float32")
        gamma = self._e_step(X)  # (batch, n_components)
        # Sample n_samples times for each measurement
        comp_idx = keras.random.categorical(
            ops.log(gamma), n_samples, seed=seed
        )  # (batch, n_samples)
        # Return as (batch, n_samples)
        return comp_idx

    def log_density(self, data, **kwargs):
        X = ops.convert_to_tensor(data, dtype="float32")
        pdf = ops.sum(self._component_pdf(X) * self.pi, axis=1)
        return ops.log(pdf)


def match_means_covariances(means, true_means, covs, true_covs):
    """Match estimated means/covs to true ones.

    Uses greedy minimal distance assignment.

    Args:
        means: Estimated means (n_components, n_features).
        true_means: True means (n_components, n_features).
        covs: Estimated covariances (n_components, n_features, n_features).
        true_covs: True covariances (n_components, n_features, n_features).

    Returns:
        means_matched: Matched estimated means.
        true_means_matched: Matched true means.
        covs_matched: Matched estimated covariances.
        true_covs_matched: Matched true covariances.
    """
    diff = ops.expand_dims(means, 1) - ops.expand_dims(true_means, 0)
    cost = ops.sqrt(ops.sum(diff**2, axis=-1))
    row_ind, col_ind = linear_sum_assignment(cost)
    means_matched = ops.take(means, row_ind, axis=0)
    true_means_matched = ops.take(true_means, col_ind, axis=0)
    covs_matched = [covs[i] for i in row_ind]
    true_covs_matched = [true_covs[j] for j in col_ind]
    return means_matched, true_means_matched, covs_matched, true_covs_matched



"""Convenience interface for loading and displaying ultrasound data.

Example usage
^^^^^^^^^^^^^^

.. code-block:: python

    import zea
    from zea.internal.setup_zea import setup_config

    config = setup_config("hf://zeahub/configs/config_camus.yaml")

    interface = zea.Interface(config)
    interface.run(plot=True)

"""

import asyncio
import time
from pathlib import Path
from typing import List

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from PIL import Image

from zea import log
from zea.config import Config
from zea.data.file import File
from zea.datapaths import format_data_path
from zea.display import to_8bit
from zea.internal.core import DataTypes
from zea.internal.viewer import (
    ImageViewerMatplotlib,
    ImageViewerOpenCV,
    filename_from_window_dialog,
    running_in_notebook,
)
from zea.io_lib import matplotlib_figure_to_numpy
from zea.ops import Pipeline
from zea.utils import keep_trying, save_to_gif, save_to_mp4


class Interface:
    """Interface for selecting / loading / processing single ultrasound images.

    Useful for inspecting datasets and single ultrasound images.

    # TODO: maybe we can refactor such that it is clear what needs to be in config.
    """

    def __init__(self, config: Config = None, verbose: bool = True, validate_file: bool = True):
        """Initialize Interface.

        Args:
            config (Config): Configuration object.
            verbose (bool): Whether to print verbose output.
            validate_file (bool): Whether to validate the file.
        """
        self.config = Config(config)
        self.verbose = verbose

        self.file = File(self.file_path)

        if validate_file:
            self.file.validate()

        # get probe and scan from file
        self.probe = self.file.probe()
        self.scan = self.file.scan(**self.config.scan)

        # initialize Pipeline
        assert "pipeline" in self.config, (
            "Pipeline not found in config, please specify pipeline in config."
        )

        self.process = Pipeline.from_config(
            self.config.pipeline,
            with_batch_dim=False,
            jit_options=None,
        )
        self.parameters = self.process.prepare_parameters(self.probe, self.scan, self.config)

        # initialize attributes for UI class
        self.data = None
        self.image = None
        self.mpl_img = None
        self.fig = None
        self.ax = None
        self.gui = None
        self.image_viewer = None

        self.plot_lib = self.config.plot.plot_lib

        if self.config.plot.headless is None:
            self.headless = False
        else:
            self.headless = self.config.plot.headless

        self.check_for_display()

        if self.plot_lib == "opencv":
            self.image_viewer = ImageViewerOpenCV(
                self.data_to_display,
                window_name=self.file.name,
                num_threads=1,
                headless=self.headless,
            )
        elif self.plot_lib == "matplotlib":
            self.image_viewer = ImageViewerMatplotlib(
                self.data_to_display,
                window_name=self.file.name,
                num_threads=1,
            )

    @property
    def dtype(self):
        """Data type of data when loaded from file."""
        return self.config.data.dtype

    @property
    def dataset_folder(self):
        """Path to dataset folder."""
        return format_data_path(self.config.data.dataset_folder, self.config.data.user)

    @property
    def file_path(self):
        """Path to data file."""
        if self.config.data.file_path:
            return self.dataset_folder / self.config.data.file_path
        else:
            return self.choose_file_path()

    @file_path.setter
    def file_path(self, value):
        """Set file path to data file."""
        self.config.data.file_path = value

    def choose_file_path(self):
        """Choose file path from window dialog."""
        if self.headless:
            raise ValueError(
                "No file path specified for data file, which is required "
                "in headless mode as window dialog cannot be opened."
            )
        filetype = "hdf5"
        log.info("Please select file from window dialog...")
        self.file_path = filename_from_window_dialog(
            f"Choose .{filetype} file",
            filetypes=((filetype, "*." + filetype),),
            initialdir=self.dataset_folder,
        )
        return self.file_path

    @property
    def data_root(self):
        """Root path to data file."""
        return Path(self.config.user.data_root)

    @dtype.setter
    def dtype(self, value):
        self.config.data.dtype = value

    @property
    def to_dtype(self):
        """Data type to convert to for display."""
        return self.config.data.to_dtype

    @to_dtype.setter
    def to_dtype(self, value):
        self.config.data.to_dtype = value

    @property
    def frame_no(self):
        """Frame number to display."""
        return self.config.data.get("frame_no")

    @frame_no.setter
    def frame_no(self, value):
        self.config.data.frame_no = value

    def check_for_display(self):
        """check if in headless mode (no monitor available)"""
        if self.headless is False:
            if matplotlib.get_backend().lower() == "agg":
                self.headless = True
                log.warning("Could not connect to display, running headless.")
        else:
            # self.plot_lib = "matplotlib"  # force matplotlib in headless mode
            matplotlib.use("agg")
            log.info("Running in headless mode as set by config.")

    def set_backend_for_notebooks(self):
        """Set backend to QtAgg if running in notebook"""
        if running_in_notebook() and not self.headless:
            matplotlib.use("QtAgg")

    def get_data(self):
        """Get data. Chosen datafile should be listed in the dataset.

        Using either file specified in config or if None, the ui window.

        Returns:
            data (np.ndarray): data array of shape (n_tx, n_el, n_ax, N_ch)
        """
        if self.verbose:
            log.info(f"Selected {log.yellow(self.file_path)}")

        # grab frame number from config or user input if not set in config
        if self.frame_no == "all":
            log.info("Will run all frames as `all` was chosen in config...")
        elif self.frame_no is None:
            if self.file.n_frames == 1:
                self.frame_no = 0
            else:
                self.frame_no = keep_trying(
                    lambda: int(input(f">> Frame number (0 / {self.file.n_frames - 1}): "))
                )

        # get data from dataset
        data = self.file.load_data(self.dtype, self.frame_no)

        return data

    def data_to_display(self, data=None):
        """Get data and convert to display to_dtype."""
        if data is None:
            self.data = self.get_data()
        else:
            self.data = data

        if self.to_dtype not in ["image", "image_sc"]:
            log.warning(
                f"Image to_dtype: {self.to_dtype} not supported for displaying data."
                "falling back to  to_dtype: `image_sc`"
            )
            self.to_dtype = "image_sc"

        # select transmits if raw or aligned data
        data_type = self.process.operations[0].input_data_type
        if data_type in [DataTypes.RAW_DATA, DataTypes.ALIGNED_DATA]:
            n_tx = self.data.shape[0]
            assert len(self.scan.selected_transmits) <= n_tx, (
                f"Number of selected transmits {len(self.scan.selected_transmits)} "
                f"exceeds number of transmits in raw data {n_tx}"
            )
            self.data = np.take(self.data, self.scan.selected_transmits, axis=0)

        inputs = {self.process.key: self.data}

        outputs = self.process(**inputs, **self.parameters)

        self.image = outputs[self.process.output_key]

        # match orientation if necessary
        if self.config.plot.fliplr:
            self.image = np.fliplr(self.image)
        # opencv requires 8 bit images
        if self.plot_lib == "opencv":
            self.image = to_8bit(self.image, self.config.data.dynamic_range)
        return self.image

    def run(self, plot=False, block=True):
        """Run ui. Will retrieve, process and plot data if set to True."""
        save = self.config.plot.save

        if self.frame_no == "all":
            if not asyncio.get_event_loop().is_running():
                asyncio.run(self.run_movie(save))
            else:
                asyncio.create_task(self.run_movie(save))

        else:
            if plot:
                self.image = self.plot(
                    save=save,
                    block=block,
                )
            else:
                self.image = self.data_to_display()

        return self.image

    def plot(
        self,
        data: np.ndarray = None,
        save: bool = False,
        block: bool = True,
    ):
        """Plot image using matplotlib or opencv.

        Args:
            save (bool): whether to save the image to disk.
            block (bool): whether to block the UI while plotting.
        Returns:
            image (np.ndarray): plotted image (grabbed from figure).
        """
        assert self.image_viewer is not None, "Image viewer not initialized."

        self.image_viewer.threading = False

        if self.plot_lib == "matplotlib":
            if self.image_viewer.fig is None:
                self._init_plt_figure()
            self.image_viewer.show(data)
            if save:
                self.save_image(self.fig)
            if not self.headless and block:
                plt.show(block=True)
            self.image = matplotlib_figure_to_numpy(self.fig)
            return self.image

        elif self.plot_lib == "opencv":
            self.image_viewer.show(data)
            if not self.headless and block:
                self.image_viewer._cv2.waitKey(0)
            self.save_image(self.image)
            return self.image

    def _init_plt_figure(self):
        figsize = (10, 10)
        if self.scan:
            extent = [
                self.scan.xlims[0] * 1e3,
                self.scan.xlims[1] * 1e3,
                self.scan.zlims[1] * 1e3,
                self.scan.zlims[0] * 1e3,
            ]
            # set figure aspect ratio to match scan
            aspect_ratio = abs(extent[1] - extent[0]) / abs(extent[3] - extent[2])
            figsize = tuple(np.array(figsize) * aspect_ratio)
        else:
            extent = None

        self.fig, self.ax = plt.subplots(figsize=figsize)

        image_range = self.config.data.dynamic_range
        imshow_kwargs = {
            "cmap": "gray",
            "vmin": image_range[0],
            "vmax": image_range[1],
            "origin": "upper",
            "extent": extent,
            "interpolation": "none",
        }
        cax_kwargs = {
            "pad": 0.05,
            "position": "right",
            "size": "5%",
        }

        self.ax.set_xlabel("Lateral Width (mm)", size=15)
        self.ax.set_ylabel("Axial length (mm)", size=15)
        self.ax.tick_params(axis="x")
        self.ax.tick_params(axis="y")

        # assign properties of fig, ax to image viewer
        self.image_viewer.imshow_kwargs = imshow_kwargs
        self.image_viewer.cax_kwargs = cax_kwargs
        self.image_viewer.fig = self.fig
        self.image_viewer.ax = self.ax

    async def run_movie(self, save: bool = False):
        """Run all frames in file in sequence"""

        log.info('Playing video, press/hold "q" while the window is active to exit...')
        self.image_viewer.threading = True
        images = await self._movie_loop(save)

        if save:
            self.save_video(images)

    async def _movie_loop(self, save: bool = False) -> List[np.ndarray]:
        """Process data and plot it in real time.

        NOTE: when plot loop is terminated by user, it will only save the shown frames.
        This is to prevent long waiting times when saving a movie (for large datasets).

        Args:
            save (bool): Whether to save the plotted images.

        Returns:
            list: A list of the plotted images.
        """
        # Initialize list of images
        images = []

        # Load correct number of frames (needs to get_data first)
        self.frame_no = 0
        self.get_data()
        n_frames = self.file.n_frames

        self.verbose = False
        try:
            while True:
                # first frame is already plotted during initialization of plotting
                start_time = time.time()
                frame_counter = 0
                self.image_viewer.frame_no = 0
                while frame_counter < n_frames:
                    if self.gui:
                        await self.gui.check_freeze()

                    await asyncio.sleep(0.01)

                    self.frame_no = frame_counter

                    if frame_counter == 0:
                        if self.plot_lib == "matplotlib":
                            if self.image_viewer.fig is None:
                                self._init_plt_figure()

                    self.image_viewer.show()

                    # set counter to frame number of image viewer (possibly not updated)
                    frame_counter = self.image_viewer.frame_no

                    # check if frame counter updated
                    if frame_counter != self.frame_no:
                        fps = frame_counter / (time.time() - start_time)
                        print(
                            f"frame {frame_counter} / {n_frames} ({fps:.2f} fps)",
                            end="\r",
                        )
                        if save and (len(images) < n_frames):
                            if self.plot_lib == "matplotlib":
                                # grab image from plt figure
                                image = matplotlib_figure_to_numpy(self.fig)
                            else:
                                image = np.array(self.image)
                            images.append(image)

                    # For opencv, show frame for 25 ms and check if "q" is pressed
                    if not self.headless:
                        if self.plot_lib == "opencv":
                            if self.image_viewer._cv2.waitKey(25) & 0xFF == ord("q"):
                                self.image_viewer.close()
                                return images
                            if self.image_viewer.has_been_closed():
                                return images
                        # For matplotlib, check if window has been closed
                        elif self.plot_lib == "matplotlib":
                            if time.sleep(0.025) and self.image_viewer.has_been_closed():
                                return images
                    # For headless mode, check if all frames have been plotted
                    if self.headless:
                        if len(images) == n_frames:
                            return images

                # clear line, frame number
                print("\x1b[2K", end="\r")

                # only loop once if in headless mode
                if self.headless:
                    return images

        except KeyboardInterrupt:
            if save:
                if len(images) > 0:
                    self.save_video(images)
            raise

    def save_image(self, fig, path=None):
        """Save image to disk.

        Args:
            fig (fig object): figure.
            path (str, optional): path to save image to. Defaults to None.

        """
        if path is None:
            if self.config.plot.tag:
                tag = "_" + self.config.plot.tag
            else:
                tag = ""

            if self.frame_no is not None:
                filename = self.file_path.stem + "-" + str(self.frame_no) + tag
            else:
                filename = self.file_path.stem + tag

            ext = f".{self.config.plot.image_extension.lstrip('.')}"

            path = Path("./figures", filename).with_suffix(ext)
            Path("./figures").mkdir(parents=True, exist_ok=True)

        if isinstance(fig, plt.Figure):
            fig.savefig(path, transparent=True)
        elif isinstance(fig, Image.Image):
            fig.save(path)
        else:
            raise ValueError(
                f"Figure is not PIL image or matplotlib figure object, got {type(fig)}"
            )

        if self.verbose:
            log.info(f"Image saved to {log.yellow(path)}")

    def save_video(self, images, path=None):
        """Save video to disk.

        Args:
            images (list): list of images.
            path (str, optional): path to save image to. Defaults to None.

        """
        if path is None:
            if self.config.plot.tag:
                tag = "_" + self.config.plot.tag
            else:
                tag = ""
            filename = self.file_path.stem + tag + "." + self.config.plot.video_extension

            path = Path("./figures", filename)
            Path("./figures").mkdir(parents=True, exist_ok=True)

        if not isinstance(images[0], np.ndarray):
            raise ValueError("Images are not numpy arrays.")

        fps = self.config.plot.fps

        if self.config.plot.video_extension == "gif":
            save_to_gif(images, path, fps=fps)
        elif self.config.plot.video_extension == "mp4":
            save_to_mp4(images, path, fps=fps)

        if self.verbose:
            log.info(f"Video saved to {log.yellow(path)}")

    def __del__(self):
        try:
            if self.image_viewer is not None:
                self.image_viewer.close()
        except Exception:
            pass
        try:
            if self.fig is not None:
                plt.close(self.fig)
        except Exception:
            pass
        try:
            if self.file is not None:
                self.file.close()
        except Exception:
            pass
